@@ -43,7 +43,8 @@ linc_close_fd (LINCConnection *cnx)
 {
 	if (cnx->priv->fd >= 0) {
 		d_printf ("Close %d\n", cnx->priv->fd);
-		close (cnx->priv->fd);
+
+		LINC_CLOSE (cnx->priv->fd);
 	}
 	cnx->priv->fd = -1;
 }
@@ -371,7 +372,7 @@ linc_connection_initiate (LINCConnection        *cnx,
  out:
 	if (!retval && fd >= 0) {
 		d_printf ("initiation failed\n");
-		close (fd);
+		LINC_CLOSE (fd);
 	}
 
 	g_free (saddr);
@@ -435,6 +436,8 @@ linc_connection_read (LINCConnection *cnx,
 #endif
 			n = read (cnx->priv->fd, buf, len);
 
+		g_assert (n <= len);
+
 		if (n < 0) {
 #ifdef LINC_SSL_SUPPORT
 			if (cnx->options & LINC_CONNECTION_SSL) {
@@ -600,6 +603,19 @@ write_data (LINCConnection *cnx, QueuedWrite *qw)
 	return bytes_written;
 }
 
+static gboolean
+linc_connection_should_block (LINCConnection      *cnx,
+			      const LINCWriteOpts *opt_write_opts)
+{
+	if (!opt_write_opts)
+		return TRUE;
+
+	if (opt_write_opts->block_on_write)
+		return TRUE;
+
+	return FALSE;
+}
+
 /**
  * linc_connection_writev:
  * @cnx: the connection to write to
@@ -622,7 +638,7 @@ linc_connection_writev (LINCConnection       *cnx,
 			const LINCWriteOpts  *opt_write_opts)
 {
 	QueuedWrite qw;
-	glong       status;
+	int         status;
 
 	/* FIXME: need an option to turn this off ? */
 	if (cnx->options & LINC_CONNECTION_NONBLOCKING) {
@@ -650,10 +666,11 @@ linc_connection_writev (LINCConnection       *cnx,
 		/* Queue data & listen for buffer space */
 		linc_watch_set_condition (cnx->priv->tag,
 					  LINC_ERR_CONDS | LINC_IN_CONDS | G_IO_OUT);
-		if (opt_write_opts &&
-		    !opt_write_opts->block_on_write) {
+
+		if (!linc_connection_should_block (cnx, opt_write_opts)) {
 			queue_flattened (cnx, qw.vecs, qw.nvecs);
 			return LINC_IO_QUEUED_DATA;
+
 		} else {
 			linc_main_iteration (TRUE);
 			goto continue_write;
