@@ -10,119 +10,15 @@
  *                 Sun Microsystems, Inc.
  */
 #include <config.h>
-
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/stat.h>
-#include <sys/time.h>
-#include <unistd.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <limits.h>
-#include <time.h>
-#include <utime.h>
-#include <errno.h>
-#include <string.h>
-#include <netinet/in.h>
-#include <netdb.h>
-
-#ifdef HAVE_NETINET_IN_H
-#include <netinet/in.h>
-#endif
-#ifdef HAVE_NETINET_TCP_H
-#include <netinet/tcp.h>
-#endif
-#ifdef HAVE_SYS_UN_H
-#include <sys/un.h>
-#endif
-#ifdef HAVE_LINUX_IRDA_H
-#include <asm/types.h>
-#include <linux/irda.h>
-#endif
-
-#include <arpa/nameser.h>
-#include <resolv.h>
-
-#if !defined (NI_MAXSERV) || !defined (NI_MAXHOST)
-#  include <netdb.h>
-#  include <sys/param.h>
-#endif
-
-#if !defined (NI_MAXHOST)
-#  define NI_MAXHOST MAXHOSTNAMELEN
-#endif
-
-#if !defined (NI_MAXSERV)
-#  define NI_MAXSERV 64
-#endif
-
+#include "linc-compat.h"
+#include <dirent.h>
 #include <linc/linc-protocol.h>
 #include <linc/linc-connection.h>
 
 #include "linc-private.h"
 #include "linc-debug.h"
 
-extern LINCProtocolInfo static_linc_protocols[];
-
-/*
- * linc_protocol_all:
- *
- * Returns a list of protocols supported by linc.
- *
- * Note: the list is terminated by a #LINCProtocolInfo with a
- *       NULL name pointer.
- *
- * Return Value: an array of #LINCProtocolInfo structures.
- */
-LINCProtocolInfo * const
-linc_protocol_all (void)
-{
-	return static_linc_protocols;
-}
-
-/*
- * linc_protocol_find:
- * @name: name of the protocol.
- *
- * Find a protocol identified by @name.
- *
- * Return Value: a pointer to a valid #LINCProtocolInfo structure if 
- *               the protocol is supported by linc, NULL otherwise.
- */
-LINCProtocolInfo * const
-linc_protocol_find (const char *name)
-{
-	int i;
-
-	for (i = 0; static_linc_protocols [i].name; i++) {
-		if (!strcmp (name, static_linc_protocols [i].name))
-			return &static_linc_protocols [i];
-	}
-
-	return NULL;
-}
-
-/*
- * linc_protocol_find_num:
- * @family: the family identifier of the protocol - i.e. AF_*
- *
- * Find a protocol identified by @family.
- *
- * Return Value: a pointer to a valid #LINCProtocolInfo structure if
- *               the protocol is supported by linc, NULL otherwise.
- */
-LINCProtocolInfo * const
-linc_protocol_find_num (const int family)
-{
-	int i;
-
-	for (i = 0; static_linc_protocols [i].name; i++) {
-		if (family == static_linc_protocols [i].family)
-			return &static_linc_protocols [i];
-	}
-
-	return NULL;
-}
+#undef LOCAL_DEBUG
 
 static char linc_tmpdir [PATH_MAX] = "";
 
@@ -154,8 +50,8 @@ make_local_tmpdir (const char *dirname)
 			if (statbuf.st_uid != getuid ())
 				g_error ("Owner of %s is not the current user\n", dirname);
 
-			if((statbuf.st_mode & (S_IRWXG|S_IRWXO))
-			   || !S_ISDIR (statbuf.st_mode))
+			if ((statbuf.st_mode & (S_IRWXG|S_IRWXO)) ||
+			    !S_ISDIR (statbuf.st_mode))
 				g_error ("Wrong permissions for %s\n", dirname);
 
 			break;
@@ -180,6 +76,8 @@ make_local_tmpdir (const char *dirname)
  * Set the temporary directory used by linc to @dir. 
  *
  * This directory is used for the creation of UNIX sockets.
+ * @dir must have the correct permissions, 0700, user owned
+ * otherwise this method will g_error.
  **/
 void
 linc_set_tmpdir (const char *dir)
@@ -191,6 +89,21 @@ linc_set_tmpdir (const char *dir)
 	make_local_tmpdir (linc_tmpdir);
 }
 
+/**
+ * linc_get_tmpdir:
+ * @void: 
+ * 
+ * Fetches the directory name used by linc to whack
+ * Unix Domain sockets into.
+ * 
+ * Return value: the g_allocated socket name.
+ **/
+char *
+linc_get_tmpdir (void)
+{
+	return g_strdup (linc_tmpdir);
+}
+
 #ifdef HAVE_SOCKADDR_SA_LEN
 #define LINC_SET_SOCKADDR_LEN(saddr, len)                     \
 		((struct sockaddr *)(saddr))->sa_len = (len)
@@ -198,17 +111,15 @@ linc_set_tmpdir (const char *dir)
 #define LINC_SET_SOCKADDR_LEN(saddr, len)
 #endif
 
-#ifdef AF_INET6
+#if defined(AF_INET6) && defined(RES_USE_INET6)
 #define LINC_RESOLV_SET_IPV6     _res.options |= RES_USE_INET6
-#define LINC_RESOLV_CLEAR_IPV6   _res.options &= ~RES_USE_INET6
 #else
 #define LINC_RESOLV_SET_IPV6
-#define LINC_RESOLV_CLEAR_IPV6
 #endif
 
 #if defined(AF_INET) || defined(AF_INET6) || defined (AF_UNIX)
-static const char *
-get_local_hostname (void)
+const char *
+linc_get_local_hostname (void)
 {
 	static char local_host[NI_MAXHOST] = { 0 };
 
@@ -221,41 +132,114 @@ get_local_hostname (void)
 	return local_host;
 }
 
+/*
+ * True if succeeded in mapping, else false.
+ */
+static gboolean
+ipv4_addr_from_addr (struct in_addr *dest_addr,
+		     guint8         *src_addr,
+		     int             src_length)
+{
+	if (src_length == 4)
+		memcpy (dest_addr, src_addr, 4);
+
+	else if (src_length == 16) {
+		int i;
+
+#ifdef LOCAL_DEBUG
+		g_warning ("Doing conversion ...");
+#endif
+
+		/* An ipv6 address, might be an IPv4 mapped though */
+		for (i = 0; i < 10; i++)
+			if (src_addr [i] != 0)
+				return FALSE;
+
+		if (src_addr [10] != 0xff ||
+		    src_addr [11] != 0xff)
+			return FALSE;
+
+		memcpy (dest_addr, &src_addr[12], 4);
+	} else
+		return FALSE;
+
+	return TRUE;
+}
+
 static gboolean
 linc_protocol_is_local_ipv46 (const LINCProtocolInfo *proto,
 			      const struct sockaddr   *saddr,
-			      socklen_t                saddr_len)
+			      LincSockLen              saddr_len)
 {
 	int i;
 	static int warned = 0;
-	static struct hostent *local_hostent = NULL;
-	
-	if (!local_hostent)
-		local_hostent = gethostbyname (get_local_hostname ());
+	static struct hostent *local_hostent;
+
+	g_assert (saddr->sa_family == proto->family);
+
+	if (!local_hostent) {
+		LINC_RESOLV_SET_IPV6;
+		local_hostent = gethostbyname (linc_get_local_hostname ());
+	}
 
 	if (!local_hostent) {
 		if (!warned++)
 			g_warning ("can't gethostbyname on '%s'",
-				   get_local_hostname ());
+				   linc_get_local_hostname ());
 		return FALSE;
 	}
 
-	if (local_hostent->h_addrtype != saddr->sa_family) {
-		if (!warned++)
-			g_warning ("FIXME: can't compare different family "
-				   "address types for locality");
+	if (!local_hostent->h_addr_list)
+		g_error ("No address for local host");
+
+	if (proto->family != AF_INET) {
+#ifdef AF_INET6
+		if (proto->family == AF_INET6 &&
+		    local_hostent->h_addrtype != AF_INET6)
+			return FALSE; /* can connect via IPv4 */
+
+		if (proto->family != AF_INET6)
+			return FALSE;
+#else
 		return FALSE;
+#endif
 	}
 
-	g_warning ("Compare vs %d matches len %d",
-		   local_hostent->h_length, saddr_len);
+	for (i = 0; local_hostent->h_addr_list [i]; i++) {
 
-	for (i = 0; i < local_hostent->h_length; i++) {
-		if (!memcmp (local_hostent->h_addr_list [i],
-			     saddr->sa_data, saddr_len))
+		if (proto->family == AF_INET) {
+			struct in_addr ipv4_addr;
+			
+			if (!ipv4_addr_from_addr (&ipv4_addr,
+						  local_hostent->h_addr_list [i],
+						  local_hostent->h_length))
+				continue;
+
+			if (!memcmp (&ipv4_addr, 
+				     &((struct sockaddr_in *)saddr)->sin_addr.s_addr, 4)) {
+#ifdef LOCAL_DEBUG
+				g_warning ("local ipv4 address");
+#endif
+				return TRUE;
+			}
+
+		}
+#ifdef AF_INET6
+		else if (!memcmp (local_hostent->h_addr_list [i],
+				  &((struct sockaddr_in6 *)saddr)->sin6_addr.s6_addr,
+				  local_hostent->h_length)) {
+#ifdef LOCAL_DEBUG
+			g_warning ("local ipv6 address");
+#endif
 			return TRUE;
+		}
+#endif
 	}
 
+#ifdef LOCAL_DEBUG
+	g_warning ("No match over all");
+#endif
+
 	return FALSE;
 }
 
@@ -279,7 +263,7 @@ static struct sockaddr *
 linc_protocol_get_sockaddr_ipv4 (const LINCProtocolInfo *proto,
 				 const char             *hostname,
 				 const char             *portnum,
-				 socklen_t              *saddr_len)
+				 LincSockLen            *saddr_len)
 {
 	struct sockaddr_in *saddr;
 	struct hostent     *host;
@@ -299,20 +283,20 @@ linc_protocol_get_sockaddr_ipv4 (const LINCProtocolInfo *proto,
 	saddr->sin_family = AF_INET;
 	saddr->sin_port   = htons (atoi (portnum));
 
+	LINC_RESOLV_SET_IPV6;
 	if (!(_res.options & RES_INIT))
 		res_init();
 
-	LINC_RESOLV_CLEAR_IPV6;
-
 	host = gethostbyname (hostname);
-	if (!host) {
+	if (!host ||
+	    !ipv4_addr_from_addr (&saddr->sin_addr,
+				  (guint8 *)host->h_addr_list [0],
+				  host->h_length)) {
 		g_free (saddr);
 		return NULL;
 	}
 
-	memcpy (&saddr->sin_addr, host->h_addr_list[0], sizeof (struct in_addr));
-
-	return (struct sockaddr *)saddr;
+	return (struct sockaddr *) saddr;
 }
 #endif /* AF_INET */
 
@@ -336,7 +320,7 @@ static struct sockaddr *
 linc_protocol_get_sockaddr_ipv6 (const LINCProtocolInfo *proto,
 				 const char             *hostname,
 				 const char             *portnum,
-				 socklen_t              *saddr_len)
+				 LincSockLen            *saddr_len)
 {
 	struct sockaddr_in6 *saddr;
 	struct hostent      *host;
@@ -360,7 +344,6 @@ linc_protocol_get_sockaddr_ipv6 (const LINCProtocolInfo *proto,
 		res_init();
 
 	LINC_RESOLV_SET_IPV6;
-
 	host = gethostbyname (hostname);
 	if (!host || host->h_addrtype != AF_INET6) {
 		g_free (saddr);
@@ -373,6 +356,7 @@ linc_protocol_get_sockaddr_ipv6 (const LINCProtocolInfo *proto,
 }
 #endif /* AF_INET6 */
 
+#ifdef AF_UNIX
 /*
  * linc_protocol_get_sockaddr_unix:
  * @proto: the #LINCProtocolInfo structure for the UNIX sockets protocol.
@@ -388,12 +372,11 @@ linc_protocol_get_sockaddr_ipv6 (const LINCProtocolInfo *proto,
  * Return Value: a pointer to a valid #sockaddr_un structure if the call 
  *               succeeds, NULL otherwise.
  */
-#ifdef AF_UNIX
 static struct sockaddr *
 linc_protocol_get_sockaddr_unix (const LINCProtocolInfo *proto,
 				 const char             *dummy,
 				 const char             *path,
-				 socklen_t              *saddr_len)
+				 LincSockLen            *saddr_len)
 {
 	struct sockaddr_un *saddr;
 	int                 pathlen;
@@ -410,22 +393,16 @@ linc_protocol_get_sockaddr_unix (const LINCProtocolInfo *proto,
 
 		gettimeofday (&t, NULL);
 		g_snprintf (buf, sizeof (buf),
-/* We'd do this all the time, but the pid is not very random,
- * and this string is truncated by some unixes */
-#ifdef CONNECTION_DEBUG
-			    "%s/l-%d-%x%x", linc_tmpdir, pid,
-#else
-			    "%s/linc-%x%x", linc_tmpdir,
-#endif
-			    (guint) (rand() ^ t.tv_sec ^ pid),
-			    (guint) (idx++ ^ t.tv_usec));
+			    "%s/linc-%x-%x-%x%x", linc_tmpdir,
+			    pid, idx++,
+			    (guint) (rand() ^ t.tv_sec),
+			    (guint) (idx ^ t.tv_usec));
 #ifdef CONNECTION_DEBUG
 		if (g_file_test (buf, G_FILE_TEST_EXISTS))
 			g_warning ("'%s' already exists !", buf);
 #endif
 		actual_path = buf;
-		}
-	else 
+	} else 
 		actual_path = (char *)path;
 
 	pathlen = strlen (actual_path);
@@ -464,7 +441,7 @@ static struct sockaddr *
 linc_protocol_get_sockaddr_irda (const LINCProtocolInfo *proto,
 				 const char             *hostname,
 				 const char             *service,
-				 socklen_t              *saddr_len)
+				 LincSockLen            *saddr_len)
 {
 	g_assert (proto->family == AF_IRDA);
 
@@ -489,8 +466,7 @@ struct sockaddr *
 linc_protocol_get_sockaddr (const LINCProtocolInfo *proto,
 			    const char             *hostname,
 			    const char             *service,
-			    socklen_t              *saddr_len)
-		   
+			    LincSockLen            *saddr_len)		   
 {
 	if (proto && proto->get_sockaddr)
 		return proto->get_sockaddr (proto, hostname, service, saddr_len);
@@ -525,9 +501,10 @@ linc_protocol_get_sockinfo_ipv46 (struct hostent  *host,
 	if (!host) {
 		const char *local_host;
 
-		if (!(local_host = get_local_hostname ()))
+		if (!(local_host = linc_get_local_hostname ()))
 			return FALSE;
 
+		LINC_RESOLV_SET_IPV6;
 		host = gethostbyname (local_host);
 	}
 
@@ -540,7 +517,7 @@ linc_protocol_get_sockinfo_ipv46 (struct hostent  *host,
 	if (portnum) {
 		gchar tmpport[NI_MAXSERV];
 
-		g_snprintf (tmpport, NI_MAXSERV, "%d", ntohs (port));
+		g_snprintf (tmpport, sizeof (tmpport), "%d", ntohs (port));
 
 		*portnum = g_strdup (tmpport);
 	}
@@ -607,6 +584,13 @@ linc_protocol_get_sockinfo_ipv4 (const LINCProtocolInfo  *proto,
  * Return Value: #TRUE if the function succeeds, #FALSE otherwise.
  */
 #ifdef AF_INET6
+
+/* FIXME: is IN6ADDR_ANY_INIT exported on Mac OS X ? */
+/* on Mac OS X 10.1 inaddr6_any isn't exported by libc */
+#ifndef in6addr_any
+	static const struct in6_addr in6addr_any = { { { 0 } } };
+#endif
+
 static gboolean
 linc_protocol_get_sockinfo_ipv6 (const LINCProtocolInfo  *proto,
 				 const struct sockaddr   *saddr,
@@ -663,7 +647,7 @@ linc_protocol_get_sockinfo_unix (const LINCProtocolInfo  *proto,
 	if (hostname) {
 		const char *local_host;
 
-		if (!(local_host = get_local_hostname ()))
+		if (!(local_host = linc_get_local_hostname ()))
 			return FALSE;
 
 		*hostname = g_strdup (local_host);
@@ -741,7 +725,7 @@ linc_protocol_get_sockinfo (const LINCProtocolInfo  *proto,
 gboolean
 linc_protocol_is_local (const LINCProtocolInfo  *proto,
 			const struct sockaddr   *saddr,
-			socklen_t                saddr_len)
+			LincSockLen              saddr_len)
 {
 	if (proto && proto->is_local)
 		return proto->is_local (proto, saddr, saddr_len);
@@ -769,7 +753,7 @@ linc_protocol_unix_destroy (int         fd,
 static gboolean
 linc_protocol_unix_is_local (const LINCProtocolInfo *proto,
 			     const struct sockaddr   *saddr,
-			     socklen_t                saddr_len)
+			     LincSockLen              saddr_len)
 {
 	return TRUE;
 }
@@ -923,7 +907,7 @@ irda_find_device (guint32  *addr,
 	}
 
  out:
-	close(fd);
+	LINC_CLOSE (fd);
 
 	return retval;
 }
@@ -970,7 +954,7 @@ irda_getaddrinfo (const char             *nodename,
 				return EAI_NONAME;
 
 			/* It's a numeric address - we need to find the hostname */
-			strcpy (hnbuf, IRDA_PREFIX);
+			g_strncpy (hnbuf, IRDA_PREFIX, IRDA_NICKNAME_MAX + IRDA_PREFIX_LEN);
 			if (irda_find_device (&sai.sir_addr, 
 					      hnbuf + IRDA_PREFIX_LEN,
 					      FALSE))
@@ -1004,7 +988,7 @@ irda_getaddrinfo (const char             *nodename,
 	retval->ai_addr = (struct sockaddr *)tptr;
 	memcpy (retval->ai_addr, &sai, sizeof (struct sockaddr_irda));
 	tptr += sizeof (struct sockaddr_irda);
-	strcpy (tptr, hnbuf);
+	g_strncpy (tptr, hnbuf, IRDA_NICKNAME_MAX + IRDA_PREFIX_LEN);
 	retval->ai_family = AF_IRDA;
 	retval->ai_socktype = SOCK_STREAM;
 	retval->ai_protocol = 0;
@@ -1017,41 +1001,42 @@ irda_getaddrinfo (const char             *nodename,
 }
 
 static int
-irda_getnameinfo(const struct sockaddr *sa, socklen_t sa_len,
-		 char *host, size_t hostlen, char *serv, size_t servlen,
-		 int flags)
+irda_getnameinfo (const struct sockaddr *sa,
+		  LincSockLen            sa_len,
+		  char                  *host,
+		  size_t                 hostlen,
+		  char                  *serv,
+		  size_t                 servlen,
+		  int                    flags)
 {
-  struct sockaddr_irda *sai = (struct sockaddr_irda *)sa;
-  gboolean got_host = FALSE;
-  /* Here, we talk to the host specified, and ask it for its name */
+	struct sockaddr_irda *sai = (struct sockaddr_irda *) sa;
+	gboolean got_host = FALSE;
+	/* Here, we talk to the host specified, and ask it for its name */
 
-  if(sa_len != sizeof(struct sockaddr_irda))
-    return -1;
+	if (sa_len != sizeof (struct sockaddr_irda))
+		return -1;
 
-  /* It doesn't seem like the sir_lsap_sel is supposed to be taken into consideration when connecting... */
-  if(!(flags & NI_NUMERICHOST))
-    {
-      char hostbuf[IRDA_NICKNAME_MAX];
-      guint32 daddr;
+	/* It doesn't seem like the sir_lsap_sel is supposed to be taken into consideration when connecting... */
+	if (!(flags & NI_NUMERICHOST)) {
+		guint32 daddr;
+		char hostbuf [IRDA_NICKNAME_MAX];
 
-      daddr = sai->sir_addr;
-      if(!irda_find_device(&daddr, hostbuf, FALSE))
-	{
-	  g_snprintf(host, hostlen, "%s", hostbuf);
-	  got_host = TRUE;
+		daddr = sai->sir_addr;
+		if (!irda_find_device (&daddr, hostbuf, FALSE)) {
+			g_snprintf (host, hostlen, "%s", hostbuf);
+			got_host = TRUE;
+		}
 	}
-    }
-  if(!got_host)
-    {
-      if(flags & NI_NAMEREQD)
-	return -1;
+	if (!got_host) {
+		if (flags & NI_NAMEREQD)
+			return -1;
 
-      g_snprintf(host, hostlen, IRDA_PREFIX "%#08x", sai->sir_addr);
-    }
+		g_snprintf (host, hostlen, IRDA_PREFIX "%#08x", sai->sir_addr);
+	}
 
-  g_snprintf(serv, servlen, "%s", sai->sir_name);
+	g_snprintf(serv, servlen, "%s", sai->sir_name);
 
-  return 0;
+	return 0;
 }
 #endif /* AF_IRDA */
 #endif /* 0 */
@@ -1068,7 +1053,7 @@ linc_protocol_destroy_cnx (const LINCProtocolInfo *proto,
 		if (proto->destroy)
 			proto->destroy (fd, host, service);
 		
-		close (fd);
+		LINC_CLOSE (fd);
 	}
 }
 
@@ -1089,9 +1074,68 @@ linc_protocol_destroy_addr (const LINCProtocolInfo *proto,
 			proto->destroy (fd, NULL, addr_un->sun_path);
 		}
 #endif
-		close (fd);
+		LINC_CLOSE (fd);
 		g_free (saddr);
 	}
 
 }
 
+/*
+ * linc_protocol_all:
+ *
+ * Returns a list of protocols supported by linc.
+ *
+ * Note: the list is terminated by a #LINCProtocolInfo with a
+ *       NULL name pointer.
+ *
+ * Return Value: an array of #LINCProtocolInfo structures.
+ */
+LINCProtocolInfo * const
+linc_protocol_all (void)
+{
+	return static_linc_protocols;
+}
+
+/*
+ * linc_protocol_find:
+ * @name: name of the protocol.
+ *
+ * Find a protocol identified by @name.
+ *
+ * Return Value: a pointer to a valid #LINCProtocolInfo structure if 
+ *               the protocol is supported by linc, NULL otherwise.
+ */
+LINCProtocolInfo * const
+linc_protocol_find (const char *name)
+{
+	int i;
+
+	for (i = 0; static_linc_protocols [i].name; i++) {
+		if (!strcmp (name, static_linc_protocols [i].name))
+			return &static_linc_protocols [i];
+	}
+
+	return NULL;
+}
+
+/*
+ * linc_protocol_find_num:
+ * @family: the family identifier of the protocol - i.e. AF_*
+ *
+ * Find a protocol identified by @family.
+ *
+ * Return Value: a pointer to a valid #LINCProtocolInfo structure if
+ *               the protocol is supported by linc, NULL otherwise.
+ */
+LINCProtocolInfo * const
+linc_protocol_find_num (const int family)
+{
+	int i;
+
+	for (i = 0; static_linc_protocols [i].name; i++) {
+		if (family == static_linc_protocols [i].family)
+			return &static_linc_protocols [i];
+	}
+
+	return NULL;
+}
